@@ -38,12 +38,13 @@
 #  part of this work is illegal and unethical regarding the effort and
 #  time spent here.
 from inspect import signature
+from types import FunctionType
 from typing import Union
 
 from lange import AP, GP
 
 from ldict.core.inspection import (
-    extract_implicit_input,
+    extract_dynamic_input,
     extract_output,
     extract_returnstr,
     extract_dictstr,
@@ -82,31 +83,80 @@ def handle_dict(data, dictlike, rnd):
 
 
 def lazify(data, output_field: Union[list, str], f, rnd, multi_output) -> Union[dict, LazyVal]:
-    """
-    Create lazy values and handle meafields.
-    """
+    """Create lazy values and handle metafields."""
+    # TODO (minor): simplify to improve readability of this function
     config, f = (f.config, f.f) if isinstance(f, AbstractLet) else ({}, f)
+    if isinstance(f, FunctionType):
+        body = extract_body(f)
+        memo = [""]
+
+        def lazy_returnstr():
+            memo[0] = extract_returnstr(body)
+            if multi_output:
+                memo[0] = extract_dictstr(memo[0])
+            return memo[0]
+
+        dynamic_input = extract_dynamic_input(lazy_returnstr)
+    else:
+        body = None
+        if not (hasattr(f, "metadata") and "input" in f.metadata and "output" in f.metadata):
+            raise Exception(f"Missing 'metadata' containing 'input' and 'output' keys for custom callable '{type(f)}'")
+        lazy_returnstr = lambda: ""
+        dynamic_input = f.metadata["input"]["dynamic"] if "dynamic" in f.metadata["input"] else []
+
+    if not dynamic_input and hasattr(f, "metadata") and "input" in f.metadata and "dynamic" in f.metadata["input"]:
+        dynamic_input = f.metadata["input"]["dynamic"]
+
     input_fields, parameters = extract_input(f)
-    body = extract_body(f)
-    returnstr = extract_returnstr(body)
-    if multi_output:
-        returnstr = extract_dictstr(returnstr)
-    for par in extract_implicit_input(returnstr):
+
+    for par in dynamic_input:
         if par not in config:  # pragma: no cover
             raise Exception(f"Parameter '{par}' value is not available:", config)
         input_fields.append(config[par])
+
+    newidx = 0
+    if hasattr(f, "metadata"):
+        step = f.metadata.copy()
+        if "id" in step:  # TODO: f id sem os valores dos parametros não conta a história toda pro oka web
+            newidx = step.pop("id")
+        for k in ["input", "output"]:
+            if k in step:
+                del step[k]
+        if "code" in f.metadata:
+            if f.metadata["code"] is ...:
+                if body is None:
+                    raise Exception(f"Cannot autofill 'metadata.code' for custom callable '{type(f)}'")
+                head = f"def f{str(signature(f))}:"
+                code = head + "\n" + body
+                f.metadata["code"] = code
+    else:
+        step = {}
+
     deps = prepare_deps(data, input_fields, parameters, config, rnd)
     if output_field == "extract":
-        explicit, meta, meta_ellipsed = extract_output(f, returnstr, deps)
+        explicit, meta, meta_ellipsed = extract_output(f, lazy_returnstr, deps)
         lazies = []
         dic = {k: LazyVal(k, f, deps, lazies) for k in explicit + meta}
         lazies.extend(dic.values())
         for metaf in meta_ellipsed:
-            if metaf == "code":
-                head = f"def f{str(signature(f))}:"
-                dic["_code"] = head + "\n" + body
-            elif metaf == "history":
-                newidx = 0
+            if metaf == "_code":
+                if hasattr(f, "metadata") and "code" in f.metadata:
+                    dic["_code"] = f.metadata["code"]
+                else:
+                    if body is None:
+                        raise Exception(f"Missing 'metadata' containing 'code' key for custom callable '{type(f)}'")
+                    head = f"def f{str(signature(f))}:"
+                    dic["_code"] = head + "\n" + body
+            elif metaf == "_function":
+                if hasattr(f, "pickle_dump"):
+                    dic["_function"] = f.pickle_dump
+                else:
+                    import dill
+
+                    dump = dill.dumps(f, protocol=5)
+                    dic["_function"] = dump
+                    f.pickle_dump = dump  # Memoize
+            elif metaf == "_history":
                 if "_history" in data:
                     last = list(data["_history"].keys())[-1]
                     if isinstance(last, int):
@@ -114,13 +164,6 @@ def lazify(data, output_field: Union[list, str], f, rnd, multi_output) -> Union[
                     dic["_history"] = data["_history"].copy()
                 else:
                     dic["_history"] = {}
-                if hasattr(f, "metadata"):
-                    step = f.metadata
-                    if "id" in f.metadata:
-                        newidx = f.metadata["id"]
-
-                else:
-                    step = {}
                 dic["_history"][newidx] = step
             else:
                 raise Exception(f"'...' is not defined for '{metaf}'.")
